package net.minecraft.server;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.World.Environment;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.Player;
import org.bukkit.event.player.PlayerPortalEvent;

public class DimensionBase {
	
	public static DimensionBase getDimByNumber(int number) {
        for(int i = 0; i < list.size(); i++) {
            DimensionBase dim = (DimensionBase)list.get(i);
            if(dim.number == number)
                return dim;
        }
        return null;
    }
	
	public static DimensionBase getDimByProvider(Class<? extends WorldProvider> worldProvider) {
        for(int i = 0; i < list.size(); i++) {
            DimensionBase dim = (DimensionBase)list.get(i);
            if(dim.worldProvider.getName().equals(worldProvider.getName()))
                return dim;
        }
        return null;
    }
	
	public WorldProvider getWorldProvider()
    {
        try {
            return (WorldProvider)worldProvider.newInstance();
        }
        catch(InstantiationException instantiationexception) { }
        catch(IllegalAccessException illegalaccessexception) { }
        return null;
    }
	
	public org.bukkit.craftbukkit.PortalTravelAgent getTeleporter()
    {
        try
        {
            if(teleporter != null)
            {
                return teleporter.newInstance();
            }
        }
        catch(InstantiationException instantiationexception) { }
        catch(IllegalAccessException illegalaccessexception) { }
        return null;
    }
	
	public static void usePortal(int toDimension, EntityPlayer entityplayer) {
		// CraftBukkit start -- Replaced the standard handling of portals with a more customised method.
        int dimension = entityplayer.dimension;
        if (dimension == toDimension)
        	toDimension = 0;
        WorldServer fromWorld = SAPI.getMinecraftInstance().getWorldServer(dimension);
        WorldServer toWorld = null;
        if (dimension < 10 || toDimension != -1)
            for (WorldServer world : SAPI.getMinecraftInstance().worlds)
                if (world.dimension == toDimension)
                    toWorld = world;
        
        DimensionBase dimOld = getDimByNumber(dimension);
        DimensionBase dimNew = getDimByNumber(toDimension);
        Loc loc = new Loc(entityplayer.locX, entityplayer.locZ);
        loc = dimNew.getDistanceScale(dimOld == null ? loc : dimOld.getDistanceScale(loc, true), false);

        Location fromLocation = new Location(fromWorld.getWorld(), entityplayer.locX, entityplayer.locY, entityplayer.locZ, entityplayer.yaw, entityplayer.pitch);
        Location toLocation = toWorld == null ? null : new Location(toWorld.getWorld(), loc.x, entityplayer.locY, loc.z, entityplayer.yaw, entityplayer.pitch);

        org.bukkit.craftbukkit.PortalTravelAgent pta = toDimension == 0 ? dimOld.getTeleporter() : dimNew.getTeleporter();
        PlayerPortalEvent event = new PlayerPortalEvent((Player) entityplayer.getBukkitEntity(), fromLocation, toLocation, pta);
        Bukkit.getServer().getPluginManager().callEvent(event);
        if (event.isCancelled() || event.getTo() == null) {
            return;
        }

        Location finalLocation = event.getTo();
        if (event.useTravelAgent()) {
            finalLocation = event.getPortalTravelAgent().findOrCreate(finalLocation);
        }
        toWorld = ((CraftWorld) finalLocation.getWorld()).getHandle();
        SAPI.getMinecraftInstance().serverConfigurationManager.moveToWorld(entityplayer, toWorld.dimension, finalLocation);
        // CraftBukkit end
	}
	
	public DimensionBase(int number, Class<? extends WorldProvider> worldProvider, Class<? extends org.bukkit.craftbukkit.PortalTravelAgent> teleporter) {
        name = "Dimension";
        this.number = number;
        this.worldProvider = worldProvider;
        this.teleporter = teleporter;
        list.add(this);
    }
	
	public Loc getDistanceScale(Loc loc, boolean goingIn) {
        return loc;
    }
	
	@SuppressWarnings("unchecked")
	public void registerEnvironment() {
		EnumHelper.addEnum(Environment.class, name.toUpperCase(), new Class[]{int.class}, new Object[]{number});
		try {
			((Map<Integer, Environment>)ModLoader.getPrivateValue(Environment.class, null, "lookup")).put(number, Environment.valueOf(name.toUpperCase()));
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

    public static List<DimensionBase> list = new ArrayList<DimensionBase>();
    public final int number;
    public final Class<? extends WorldProvider> worldProvider;
    public final Class<? extends org.bukkit.craftbukkit.PortalTravelAgent> teleporter;
    public String name;
    
    static {
    	new DimensionOverworld();
    	MinecraftServer mc = SAPI.getMinecraftInstance();
    	if (mc.propertyManager == null ? new PropertyManager(mc.options).getBoolean("allow-nether", true) : mc.propertyManager.getBoolean("allow-nether", true))
    		new DimensionNether();
    }
    
    public static class EnumHelper {
        
        private static Object reflectionFactory      = null;
        private static Method newConstructorAccessor = null;
        private static Method newInstance            = null;
        private static Method newFieldAccessor       = null;
        private static Method fieldAccessorSet       = null;
        private static boolean isSetup               = false;
        
        private static void setup()
        {
            if (isSetup)
                return;

            try {
                Method getReflectionFactory = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("getReflectionFactory");
                reflectionFactory      = getReflectionFactory.invoke(null);
                newConstructorAccessor = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("newConstructorAccessor", Constructor.class);
                newInstance            = Class.forName("sun.reflect.ConstructorAccessor").getDeclaredMethod("newInstance", Object[].class);
                newFieldAccessor       = Class.forName("sun.reflect.ReflectionFactory").getDeclaredMethod("newFieldAccessor", Field.class, boolean.class);
                fieldAccessorSet       = Class.forName("sun.reflect.FieldAccessor").getDeclaredMethod("set", Object.class, Object.class);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }

            isSetup = true;
        }
        
        private static Object getConstructorAccessor(Class<?> enumClass, Class<?>[] additionalParameterTypes) throws Exception {
            Class<?>[] parameterTypes = new Class[additionalParameterTypes.length + 2];
            parameterTypes[0] = String.class;
            parameterTypes[1] = int.class;
            System.arraycopy(additionalParameterTypes, 0, parameterTypes, 2, additionalParameterTypes.length);
            return newConstructorAccessor.invoke(reflectionFactory, enumClass.getDeclaredConstructor(parameterTypes));
        }
        
        private static < T extends Enum<? >> T makeEnum(Class<T> enumClass, String value, int ordinal, Class<?>[] additionalTypes, Object[] additionalValues) throws Exception {
            Object[] parms = new Object[additionalValues.length + 2];
            parms[0] = value;
            parms[1] = Integer.valueOf(ordinal);
            System.arraycopy(additionalValues, 0, parms, 2, additionalValues.length);
            return enumClass.cast(newInstance.invoke(getConstructorAccessor(enumClass, additionalTypes), new Object[] {parms}));
        }
        
        public static void setFailsafeFieldValue(Field field, Object target, Object value) throws Exception {
            field.setAccessible(true);
            Field modifiersField = Field.class.getDeclaredField("modifiers");
            modifiersField.setAccessible(true);
            modifiersField.setInt(field, field.getModifiers() & ~Modifier.FINAL);
            Object fieldAccessor = newFieldAccessor.invoke(reflectionFactory, field, false);
            fieldAccessorSet.invoke(fieldAccessor, target, value);
        }
        
        private static void blankField(Class<?> enumClass, String fieldName) throws Exception {
            for (Field field : Class.class.getDeclaredFields())
                if (field.getName().contains(fieldName)) {
                    field.setAccessible(true);
                    setFailsafeFieldValue(field, enumClass, null);
                    break;
                }
        }
        
        private static void cleanEnumCache(Class<?> enumClass) throws Exception
        {
            blankField(enumClass, "enumConstantDirectory");
            blankField(enumClass, "enumConstants");
        }
        
        @SuppressWarnings("unchecked")
        public static <T extends Enum<? >> T addEnum(Class<T> enumType, String enumName, Class<?>[] paramTypes, Object[] paramValues) {
            if (!isSetup)
                setup();

            Field valuesField = null;
            Field[] fields = enumType.getDeclaredFields();
            
            for (Field field : fields) {
                String name = field.getName();
                if (name.equals("$VALUES") || name.equals("ENUM$VALUES")) { //Added 'ENUM$VALUES' because Eclipse's internal compiler doesn't follow standards
                    valuesField = field;
                    break;
                }
            }

            int flags = Modifier.PRIVATE | Modifier.STATIC | Modifier.FINAL | 0x1000 /*SYNTHETIC*/;
            if (valuesField == null) {
                String valueType = String.format("[L%s;", enumType.getName().replace('.', '/'));
        
                for (Field field : fields)
                    if ((field.getModifiers() & flags) == flags &&
                         field.getType().getName().replace('.', '/').equals(valueType)) { //Apparently some JVMs return .'s and some don't..
                        valuesField = field;
                        break;
                    }
            }

            if (valuesField == null)
                return null;

            valuesField.setAccessible(true);

            try {
                T[] previousValues = (T[])valuesField.get(enumType);
                List<T> values = new ArrayList<T>(Arrays.asList(previousValues));
                T newValue = (T)makeEnum(enumType, enumName, values.size(), paramTypes, paramValues);
                values.add(newValue);
                setFailsafeFieldValue(valuesField, null, values.toArray((T[]) Array.newInstance(enumType, 0)));
                cleanEnumCache(enumType);

                return newValue;
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        
        static {
            if (!isSetup)
                setup();
        }
    }
}
